#
# Copyright 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from ironicclient.common import base


class Event(base.Resource):
    def __repr__(self) -> str:
        return "<Event: %s>" % self.name


class EventManager(base.CreateManager[Event]):
    resource_class: type[Event] = Event
    _creation_attributes: list[str] = ['events']
    _resource_name: str = 'events'
