#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import builtins
from typing import Any

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


class Runbook(base.Resource):
    def __repr__(self) -> str:
        return "<Runbook %s>" % self._info


class RunbookManager(base.CreateManager[Runbook]):
    resource_class: type[Runbook] = Runbook
    _creation_attributes: list[str] = [
        'description', 'extra', 'name', 'owner', 'public', 'steps',
        'traits', 'uuid',
    ]
    _resource_name: str = 'runbooks'

    def get(
        self,
        runbook_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        return self._get(
            resource_id=runbook_id,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def delete(
        self,
        runbook_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        return self._delete(
            resource_id=runbook_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def update(
        self,
        runbook_id: str,
        patch: list[dict[str, Any]],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource:
        return self._update(
            resource_id=runbook_id,
            patch=patch,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        limit: int | None = None,
        marker: str | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        project: str | None = None,
        public: bool | None = None,
    ) -> list[Runbook]:
        """Retrieve a list of runbooks.

        :param marker: Optional, the UUID of a deploy template, eg the last
                       template from a previous result set. Return the next
                       result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of runbooks to return.
            2) limit == 0, return the entire list of runbooks.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about runbooks.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of runbooks.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir,
            fields, detail=detail, public=public,
            project=project)
        path = ''
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "runbooks",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "runbooks",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def get_traits(
        self,
        runbook_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> builtins.list[str]:
        """Get traits for a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = f"{runbook_ident}/traits"
        return self._list_primitives(
            self._path(path), 'traits',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def add_trait(
        self,
        runbook_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Runbook | None:
        """Add a trait to a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param trait: trait to add to the runbook.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits/%s" % (runbook_ident, trait)
        return self._update(
            resource_id=path,
            patch=None,
            method='PUT',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def remove_trait(
        self,
        runbook_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove a trait from a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param trait: trait to remove from the runbook.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits/%s" % (runbook_ident, trait)
        self._delete(
            resource_id=path,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def remove_all_traits(
        self,
        runbook_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove all traits from a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits" % runbook_ident
        self._delete(
            resource_id=path,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)
